#! /usr/bin/env python

"""
Set up OF-DPA groups and flows configuring L2 forwarding between two physical ports. Flows
in both directions are configured by this script.

This script invokes OF-DPA API services via RPC. The RPC calls are served by the ofdpa
process running on the switch.

"""
from OFDPA_python import *

def main():
    rc = ofdpaClientInitialize("OFDPA_example")
    if rc == OFDPA_E_NONE:
        # variables used to set up bridging scenario
        vlanId = 10
        firstPhysicalPort = 3
        secondPhysicalPort = 4
        firstMacAddress = "00:00:00:00:00:01"
        secondMacAddress = "00:00:00:00:00:02"

        groupId1_p = new_uint32_tp()
        groupId2_p = new_uint32_tp()

        uint32_tp_assign(groupId1_p, 0)
        uint32_tp_assign(groupId2_p, 0)

        # set up L2 Interface groups on physical ports
        l2IfaceGroupEntry1 = ofdpaGroupEntry_t()
        l2IfaceGroupEntry2 = ofdpaGroupEntry_t()
        l2IfaceGroupBucket1 = ofdpaGroupBucketEntry_t()
        l2IfaceGroupBucket2 = ofdpaGroupBucketEntry_t()

        ofdpaGroupTypeSet(groupId1_p, OFDPA_GROUP_ENTRY_TYPE_L2_INTERFACE)
        ofdpaGroupVlanSet(groupId1_p, vlanId)
        ofdpaGroupPortIdSet(groupId1_p, firstPhysicalPort)
        l2IfaceGroupEntry1.groupId = uint32_tp_value(groupId1_p)

        l2IfaceGroupBucket1.groupId = l2IfaceGroupEntry1.groupId
        l2IfaceGroupBucket1.bucketIndex = 0
        l2IfaceGroupBucket1.bucketData.l2Interface.outputPort = firstPhysicalPort
        l2IfaceGroupBucket1.bucketData.l2Interface.popVlanTag = 1

        ofdpaGroupAdd(l2IfaceGroupEntry1)
        ofdpaGroupBucketEntryAdd(l2IfaceGroupBucket1)

        ofdpaGroupTypeSet(groupId2_p, OFDPA_GROUP_ENTRY_TYPE_L2_INTERFACE)
        ofdpaGroupVlanSet(groupId2_p, vlanId)
        ofdpaGroupPortIdSet(groupId2_p, secondPhysicalPort)
        l2IfaceGroupEntry2.groupId = uint32_tp_value(groupId2_p)

        l2IfaceGroupBucket2.groupId = l2IfaceGroupEntry2.groupId
        l2IfaceGroupBucket2.bucketData.l2Interface.outputPort = secondPhysicalPort

        ofdpaGroupAdd(l2IfaceGroupEntry2)
        ofdpaGroupBucketEntryAdd(l2IfaceGroupBucket2)

        # set up VLAN Filtering flow entries

        #
        # Have to create a flow for tagged packets before creating an untagged flow.
        #
        vlanFlowEntry1 = ofdpaFlowEntry_t()
        ofdpaFlowEntryInit(OFDPA_FLOW_TABLE_ID_VLAN, vlanFlowEntry1)
        vlanFlowEntry1.flowData.vlanFlowEntry.gotoTableId = OFDPA_FLOW_TABLE_ID_TERMINATION_MAC
        vlanFlowEntry1.flowData.vlanFlowEntry.match_criteria.inPort = firstPhysicalPort
        vlanFlowEntry1.flowData.vlanFlowEntry.match_criteria.vlanId = (OFDPA_VID_PRESENT | vlanId)
        vlanFlowEntry1.flowData.vlanFlowEntry.match_criteria.vlanIdMask = (OFDPA_VID_PRESENT | OFDPA_VID_EXACT_MASK)

        ofdpaFlowAdd(vlanFlowEntry1)

        #
        # Now create the flow for untagged packets.
        #
        vlanFlowEntry1.flowData.vlanFlowEntry.match_criteria.vlanId = 0
        vlanFlowEntry1.flowData.vlanFlowEntry.setVlanIdAction = 1
        vlanFlowEntry1.flowData.vlanFlowEntry.newVlanId = (OFDPA_VID_PRESENT | vlanId)
        ofdpaFlowAdd(vlanFlowEntry1)

        vlanFlowEntry2 = ofdpaFlowEntry_t()
        ofdpaFlowEntryInit(OFDPA_FLOW_TABLE_ID_VLAN, vlanFlowEntry2)
        vlanFlowEntry2.flowData.vlanFlowEntry.gotoTableId = OFDPA_FLOW_TABLE_ID_TERMINATION_MAC
        vlanFlowEntry2.flowData.vlanFlowEntry.match_criteria.inPort = secondPhysicalPort
        vlanFlowEntry2.flowData.vlanFlowEntry.match_criteria.vlanId = (OFDPA_VID_PRESENT | vlanId)
        vlanFlowEntry2.flowData.vlanFlowEntry.match_criteria.vlanIdMask = (OFDPA_VID_PRESENT | OFDPA_VID_EXACT_MASK)

        ofdpaFlowAdd(vlanFlowEntry2)

        # set up first Bridging flow entry
        bridgingFlowEntry1 = ofdpaFlowEntry_t()
        ofdpaFlowEntryInit(OFDPA_FLOW_TABLE_ID_BRIDGING, bridgingFlowEntry1)
        bridgingFlowEntry1.flowData.bridgingFlowEntry.gotoTableId = OFDPA_FLOW_TABLE_ID_ACL_POLICY
        bridgingFlowEntry1.flowData.bridgingFlowEntry.groupID = l2IfaceGroupEntry2.groupId
        bridgingFlowEntry1.flowData.bridgingFlowEntry.match_criteria.vlanId = (OFDPA_VID_PRESENT | vlanId)
        bridgingFlowEntry1.flowData.bridgingFlowEntry.match_criteria.vlanIdMask = (OFDPA_VID_PRESENT | OFDPA_VID_EXACT_MASK)
        MACAddress_set(bridgingFlowEntry1.flowData.bridgingFlowEntry.match_criteria.destMac, firstMacAddress)
        MACAddress_set(bridgingFlowEntry1.flowData.bridgingFlowEntry.match_criteria.destMacMask, "ff:ff:ff:ff:ff:ff")

        ofdpaFlowAdd(bridgingFlowEntry1)

        # set up second Bridging flow entry
        bridgingFlowEntry2 = ofdpaFlowEntry_t()
        ofdpaFlowEntryInit(OFDPA_FLOW_TABLE_ID_BRIDGING, bridgingFlowEntry2)
        bridgingFlowEntry2.flowData.bridgingFlowEntry.gotoTableId = OFDPA_FLOW_TABLE_ID_ACL_POLICY
        bridgingFlowEntry2.flowData.bridgingFlowEntry.groupID = l2IfaceGroupEntry1.groupId
        bridgingFlowEntry2.flowData.bridgingFlowEntry.match_criteria.vlanId = (OFDPA_VID_PRESENT | vlanId)
        bridgingFlowEntry2.flowData.bridgingFlowEntry.match_criteria.vlanIdMask = (OFDPA_VID_PRESENT | OFDPA_VID_EXACT_MASK)
        MACAddress_set(bridgingFlowEntry2.flowData.bridgingFlowEntry.match_criteria.destMac, secondMacAddress)
        MACAddress_set(bridgingFlowEntry2.flowData.bridgingFlowEntry.match_criteria.destMacMask, "ff:ff:ff:ff:ff:ff")

        ofdpaFlowAdd(bridgingFlowEntry2)

        # clean up
        delete_uint32_tp(groupId1_p)
        delete_uint32_tp(groupId2_p)

    else:
        print "Unable to initialize."


if __name__ == '__main__': main()
